-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local M = {}

M.type = "main"
M.relevantDevice = nil

local nodes = { L = {}, R = {} , B = {}}
local hotspot = {x = 0, y = 0, r_max = 0, r_min=0 , v_max = 0 , v_min = 0 , dist = 0, life = 0}
local thermal = {v = 0, r = 0 , alt = 0, life = 0}
local coldspot = {x = 0, y = 0}
local spawn_time = 0
local objpos = 0
local objvel = 0
local rot = 0
local dirVector = 0 
local jitters = {stmp={},a={},p={},amp = {} , per = {}, off={}}
local hotheight = 0
local wingpos = { L = 0, R = 0}
local info = {condition = 0, maxalt = 0, distance = 0}

local function sign(x)
  if x >= 0 then
	return 1
  end
  if x< 0 then 
	return -1
  end
end

local function inclamp(x)
	if math.abs(x) < 1 then
		return sign(x)
	end
	return x
end

local function getdistance(spot,pos)
	return math.sqrt(math.pow((pos.x-spot.x),2)+math.pow((pos.y-spot.y),2))
end

local function set_jitter(id)
	jitters.a[id] = (math.random(0.5,1))*jitters.amp[id]
	jitters.p[id] = math.random(6,15)*jitters.per[id]
	jitters.stmp[id] = 0
end

local function init_jitter(id,amp,per,off)
	jitters.amp[id] = amp
	jitters.per[id] = per
	jitters.off[id] = off
	jitters.a[id] = (math.random(0.5,1))*jitters.amp[id]
	jitters.p[id] = math.random(6,15)*jitters.per[id]
	jitters.stmp[id] = 0
end

local function spawn_hotspot()
	local rand = math.random(math.max(info.condition,0),math.min(info.condition,0)+100)/100
	objpos = obj:getPosition()
	hotspot.x = objpos.x + (math.random(-1,1))*info.distance
	hotspot.y = objpos.y + (math.random(-1,1))*info.distance
	hotspot.v_max = 10  * (rand*0.6+0.4)
	hotspot.v_min = 4 * (rand*0.6+0.5)
	hotspot.r_max = 300 * (rand*0.85+0.15)
	hotspot.r_min = 0.25*hotspot.r_max
	hotspot.life = 10*60*math.pow(rand,2)
	thermal.alt = (info.maxalt-info.minalt) * math.random() + info.minalt
	hotspot.dist = getdistance(hotspot,objpos)
	coldspot.x = hotspot.x + 1.0*hotspot.r_max * inclamp(math.random(-2,2))
	coldspot.y = hotspot.y + 1.0*hotspot.r_max * inclamp(math.random(-2,2))
	init_jitter(0,0.4,0.2,1)
	if rand > 0.4 then
		init_jitter(1,rand*0.2,0.2,1.05)
	else
		init_jitter(1,2.5,0.2,2)
	end
	init_jitter(2,1-rand*0.7,0.4,1)
	init_jitter(3,1-rand*0.7,0.4,1)
	init_jitter(4,1-rand*0.7,0.4,1)
	init_jitter(5,1-rand*0.7,0.4,1)
	print("New Hotspot has been spawned")
	print("Goodness: 			" .. tostring(rand))
	print("Max Radius :         " .. tostring(hotspot.r_max))
	print("Min Radius :         " .. tostring(hotspot.r_min))
	print("Max Vertical Speed:  " .. tostring(hotspot.v_max))
	print("Min Vertical Speed:  " .. tostring(hotspot.v_min))
	print("Start height:        " .. tostring(thermal.alt))
	print("Distance to Thermal: " .. tostring(hotspot.dist))
end

local function updateTM()
  objpos = obj:getPosition()
  dirVector = obj:getDirectionVector()
  local dirVectorUp = obj:getDirectionVectorUp()
  local cross3d = vec3(dirVector.y*dirVectorUp.z - dirVectorUp.y*dirVector.z , dirVector.z*dirVectorUp.x - dirVector.x*dirVectorUp.z , dirVector.x * dirVectorUp.y - dirVector.y * dirVectorUp.x)
  wingpos.L = -8*cross3d + objpos
  wingpos.R = 8*cross3d + objpos
  local rollDeg = (dirVectorUp.x * -dirVector.y + dirVectorUp.y * dirVector.x) * 57.2958
  local pitchDeg = dirVector.z * 57.2958
  local yawDeg = obj:getDirection() * 57.2958
  rot = vec3(pitchDeg,rollDeg,yawDeg)
end

local function init(jbeamData)
	info.condition = jbeamData.condition
	info.distance = jbeamData.distance
	info.maxalt = jbeamData.maxalt
	info.minalt = jbeamData.minalt
	print("Simulation initialized Thermal condition at " .. tostring(info.condition) .. "%, spawn distance: " .. tostring(info.distance) .. "m, cloud limit " .. tostring(jbeamData.maxalt) .. "m and min spaw altitude: " .. tostring(jbeamData.minalt) .. "m")
	for k, node in pairs (v.data.nodes) do
		if node.name == "wing29" then nodes.R[0] = k end
		if node.name == "wing21" then nodes.R[1] = k end
		if node.name == "wing15" then nodes.R[2] = k end
		if node.name == "wing6" then nodes.R[3] = k end
		if node.name == "wing11" then nodes.R[4] = k end
		if node.name == "wing9" then nodes.R[5] = k end
		if node.name == "wing28" then nodes.L[0] = k end
		if node.name == "wing20" then nodes.L[1] = k end
		if node.name == "wing14" then nodes.L[2] = k end
		if node.name == "wing1" then nodes.L[3] = k end
		if node.name == "wing10" then nodes.L[4] = k end
		if node.name == "wing8" then nodes.L[5] = k end
		if node.name == "body28" then nodes.B[0] = k end
		if node.name == "body27" then nodes.B[1] = k end
		if node.name == "body75" then nodes.B[2] = k end
		if node.name == "body74" then nodes.B[3] = k end
	end
	spawn_hotspot()
	updateTM()
	electrics.values['head'] = 0
	electrics.values['rolldeg'] = 0
end

local function onReset()
	spawn_hotspot()
end

local function jitter(dt,id)
	jitters.stmp[id] = jitters.stmp[id] + dt*jitters.p[id]
	if jitters.stmp[id] >= math.pi then
		set_jitter(id)
	end
	return (jitters.off[id]-jitters.a[id]*math.pow(math.sin(jitters.stmp[id]),2))
end

local function moveThermal(dt)
	thermal.alt = thermal.alt + dt * (hotspot.v_min+(hotspot.v_max-hotspot.v_min)/info.maxalt*thermal.alt)
	thermal.life = thermal.life + dt

end

local function calc_shape(height,dt)
	if height > 2/3*info.maxalt then
		thermal.v = math.max(hotspot.v_max*(3-3*height/info.maxalt),0)
	else
		thermal.v = hotspot.v_min+(hotspot.v_max-hotspot.v_min)/info.maxalt * 3/2 * height
	end
	thermal.r = jitter(dt,0)*hotspot.r_min+(hotspot.r_max-hotspot.r_min)/info.maxalt * height
	if height > thermal.alt then
		thermal.v = thermal.v *math.max(10+thermal.alt-height,0)/10
	end
	if thermal.life > hotspot.life then
		thermal.v = thermal.v * math.max(hotspot.life + 100 - thermal.life,0)/100
	end
	thermal.v = jitter(dt,1)*thermal.v
end

local function calc_air(dist,dt,id)
	local v_air = math.abs((1-rot.x/90)*(1-rot.y/90))*thermal.v
	
	local airspeed = electrics.values.airspeed*3.6
	if dist > thermal.r then
		v_air = v_air * (1/(1+0.03*math.pow(1/thermal.r,2)*math.pow(dist-thermal.r,4)))*jitter(dt,id)
	end
	if math.abs(v_air) < 0.01 then
		v_air = 0
	end
	if airspeed < 40 then
		v_air = v_air * airspeed/40
	end
	return v_air
end

local function heading()
	local norm = vec3(objpos.x-hotspot.x,objpos.y-hotspot.y,0):length()
	local norm2 = vec3(dirVector.x,dirVector.y,0):length()
	local angle = math.atan2(dirVector.y/norm2,dirVector.x/norm2)-math.atan2((hotspot.y-objpos.y)/norm,(hotspot.x-objpos.x)/norm)
	electrics.values['head'] = angle*57.2958
	electrics.values['rolldeg'] = rot.y
end


local function updateGFX(dt)
	if (info.condition > -95) then
		updateTM()
		heading()
		moveThermal(dt)
		calc_shape(objpos.z,dt)
		local distR_u = getdistance(hotspot,wingpos.R)
		local distL_u = getdistance(hotspot,wingpos.L)
		local distR_d = getdistance(coldspot,wingpos.R)
		local distL_d = getdistance(coldspot,wingpos.L)
		local v_air_R = calc_air(distR_u,dt,2)-1.3*calc_air(distR_d,dt,3)
		local v_air_L = calc_air(distL_u,dt,4)-1.3*calc_air(distL_d,dt,5)
		if distR_u > math.max(2*(info.distance),2*thermal.r) or thermal.life > hotspot.life +100 then
			spawn_hotspot()
		end
		local v_air = (v_air_R+v_air_L)/2
		local factor = -185000*dt
		--Moves the Aircraft upwards
		obj:applyForce(nodes.R[0], nodes.R[1] , v_air_R*factor*0.15)
		obj:applyForce(nodes.L[0], nodes.L[1] , v_air_L*factor*0.15)
		obj:applyForce(nodes.R[2], nodes.R[3] , v_air_R*factor*0.35)
		obj:applyForce(nodes.L[2], nodes.L[3] , v_air_L*factor*0.35)
		obj:applyForce(nodes.B[0], nodes.B[1] , v_air*factor*0.5)
		obj:applyForce(nodes.B[2], nodes.B[3] , v_air*factor*0.6)
		obj:applyForce(nodes.R[4], nodes.R[5] , v_air_R*factor*2)
		obj:applyForce(nodes.L[4], nodes.L[5] , v_air_L*factor*2)

	end
end

-- public interface
M.init    = init
M.updateGFX = updateGFX

--Mandatory main controller API
M.shiftUp = nop
M.shiftDown = nop
M.shiftToGearIndex = nop
M.cycleGearboxModes = nop
M.setGearboxMode = nop
M.setStarter = nop
M.setFreeze = nop
M.sendTorqueData = nop
-------------------------------

return M